#include <sys/stat.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <math.h>
#include "CL/opencl.h"
#include "AOCLUtils/aocl_utils.h"
#include "linear_classifier.h"


using namespace aocl_utils;

#define IMG_SIZE 		28*28
#define NUM_IMAGES 	500 //1000
#define TEST_NUM 		8

unsigned int correct_count = 0;

// OpenCL runtime configuration
cl_platform_id platform = NULL;
unsigned num_devices = 0;
scoped_array<cl_device_id> device; // num_devices elements
cl_context context = NULL;
scoped_array<cl_command_queue> queue; // num_devices elements
cl_program program = NULL;
scoped_array<cl_kernel> kernel; // num_devices elements

scoped_array<cl_mem> input_x_buf;
scoped_array<cl_mem> output_buf;

float* input_x[NUM_IMAGES];
float output[Y_y];

// Problem data
unsigned W_height = W_y;
unsigned W_width  = W_x;
unsigned X_height = X_y;
unsigned X_width  = X_x;
unsigned Y_height = Y_y;
unsigned Y_width  = Y_x;

// Open MNIST dataset file
std::ifstream file("../../mnist_dataset/separated/data8", std::ios::in | std::ios::binary);

// Open time log file
std::ofstream log_file("logs/time_log_std.log");

// Tot time
cl_ulong tot_kernel_time_ns = 0;
double total_time;


// Function prototypes
void read_mnist();
bool init_opencl();
void run();
void verify();
void cleanup();

int main(void) {
	// Read the MNIST data
	read_mnist();

	// Initialize OpenCL.
	if (!init_opencl()) {
		return -1;
	}

	// Run the kernel
	run();

	// Free the resources allocated
	cleanup();

	// Calculate accuracy
	float accuracy = ((float)correct_count/NUM_IMAGES)*100;
	printf("Accuracy: %f\n", accuracy);
	return 0;
}

/////////// HELPER FUNCTIONS //////////

// Reads NUM_IMAGES from MNIST file
void read_mnist() {
	for (int i = 0; i < NUM_IMAGES; i++) {
    char buffer[IMG_SIZE];

    if (file.is_open()){
      file.read(buffer, IMG_SIZE); // values are 0-255
    } else {
  		printf("Could not read file\n");
  		file.close();
  	}

		// 64 byte aligned malloc for DMA
    input_x[i] = (float*)alignedMalloc(sizeof(float)*IMG_SIZE+1);

		// Normalize the inputs
		for (int j=0; j < IMG_SIZE; j++){
			input_x[i][j] = ((unsigned char)buffer[j])/255.0;
		}
    input_x[i][IMG_SIZE] = 1;
  }
}

// Initializes the OpenCL objects
bool init_opencl() {
  cl_int status;

  printf("Initializing OpenCL\n");

  if (!setCwdToExeDir()) {
    return false;
  }

  // Get the OpenCL platform
  platform = findPlatform("Altera");
  if (platform == NULL) {
    printf("ERROR: Unable to find Altera OpenCL platform.\n");
  }

  // Query the available OpenCL device.
  device.reset(getDevices(platform, CL_DEVICE_TYPE_ALL, &num_devices));
  printf("Platform: %s\n", getPlatformName(platform).c_str());
  printf("Using %d device(s)\n", num_devices);
  for (unsigned i = 0; i < num_devices; ++i) {
    printf("  %s\n", getDeviceName(device[i]).c_str());
  }

  // Create the context
  context = clCreateContext(NULL, num_devices, device, &oclContextCallback, NULL, &status);
  checkError(status, "Failed to create context");

  // Create the program for all device. Use the first device as the
  // representative device (assuming all device are of the same type)
  std::string binary_file = getBoardBinaryFile("linear_classifier", device[0]);
  printf("Using AOCX: %s\n", binary_file.c_str());
  program = createProgramFromBinary(context, binary_file.c_str(), device , num_devices);

  // Build the program that was just created
  status = clBuildProgram(program, 0, NULL, "", NULL, NULL);
  checkError(status, "Failed to build program");

  // Create per-device objects
  queue.reset(num_devices);
  kernel.reset(num_devices);
	input_x_buf.reset(num_devices*NUM_IMAGES);
	output_buf.reset(num_devices*NUM_IMAGES);

  const unsigned num_block_rows = Y_height;

  for (unsigned i = 0; i < num_devices; ++i) {
    // Command queue
    queue[i] = clCreateCommandQueue(context, device[i], CL_QUEUE_PROFILING_ENABLE, &status);
    checkError(status, "Failed to create command queue");

    // Kernel.
    const char *kernel_name = "linear_classifier";
    kernel[i] = clCreateKernel(program, kernel_name, &status);
    checkError(status, "Failed to create kernel");

		for (unsigned j = 0; j < NUM_IMAGES; j++){
			// Create input and output buffers
			input_x_buf[j] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_BANK_2_ALTERA,
					X_height * X_width * sizeof(float), NULL, &status);
			checkError(status, "Failed to create buffer for input x");

			output_buf[j] = clCreateBuffer(context, CL_MEM_WRITE_ONLY | CL_MEM_BANK_1_ALTERA,
	       Y_height * Y_width * sizeof(float), NULL, &status);
	    checkError(status, "Failed to create buffer for output");

			// Fill input buffer with all input images
			status = clEnqueueWriteBuffer(queue[0], input_x_buf[j], CL_TRUE, // CL_FALSE
					0, X_width * X_height * sizeof(float), input_x[j], 0, NULL, NULL);
			checkError(status, "Failed to transfer input X");
		}
  }

  return true;
}

void run() {
  cl_int status;

  // Launch kernels.

  // This is the portion of time that we'll be measuring for throughput
  // benchmarking.
  scoped_array<cl_event> kernel_event(num_devices);
  const double start_time = getCurrentTimestamp();

	// Enqueue kernel.
	// Use a global work size corresponding to the size of the output matrix.
	// Each work-item computes the result for one value of the output matrix,
	// so the global work size has the same dimensions as the output matrix.
	//
	// Events are used to ensure that the kernel is not launched until
	// the writes to the input buffers have completed.
	const size_t global_work_size[1] = {10};
	const size_t local_work_size[1]  = {10};

	printf("Launching for device %d (global size: %zd)\n", 0, global_work_size[0]);

	for(unsigned i = 0; i < NUM_IMAGES; i++) {
		// Set correct output buffer for argument
		status = clSetKernelArg(kernel[0], 0, sizeof(cl_mem), &output_buf[i]);
		checkError(status, "Failed to set argument %d", 0);

		// Set correct input buffer
		status = clSetKernelArg(kernel[0], 1, sizeof(cl_mem), &input_x_buf[i]);
		checkError(status, "Failed to set argument %d", 1);

		// Execute the kernel
		status = clEnqueueNDRangeKernel(queue[0], kernel[0], 1, NULL,
				global_work_size, local_work_size, 0, NULL, &kernel_event[0]);
		checkError(status, "Failed to launch kernel");
	}

	// Wait for all kernels to finish.
	clWaitForEvents(num_devices, kernel_event);

	// Get kernel times using the OpenCL event profiling API.
	cl_ulong time_ns = getStartEndTime(kernel_event[0]);
	//printf("Kernel time (device %d): %0.3f ms\n", 0, double(time_ns) * 1e-6);
	tot_kernel_time_ns += time_ns;

	const double end_time = getCurrentTimestamp();
	total_time = end_time - start_time;

	// Wall-clock time taken.
	printf("\nTime: %0.3f ms\n", total_time * 1e3);

	// Write to timing log file
	log_file << "Total time: " << std::setprecision(3) <<  total_time * 1e3 << "ms\n";
	log_file << "\nTotal kernel time: " << std::setprecision(3) <<  tot_kernel_time_ns*NUM_IMAGES * 1e-6 << "ms\n";
	log_file << "\nKernel time: " << std::setprecision(3) <<  tot_kernel_time_ns * 1e-6 << "ms\n";
	log_file.close();

	// Compute the throughput (GFLOPS).
	// There are Y_width * Y_height output values, with each value
	// computed using W_width multiplies and adds.
	const float flops = (float)(2.0f * Y_width * Y_height * W_width * NUM_IMAGES / total_time);
	printf("\nThroughput: %0.2f GFLOPS\n\n", flops * 1e-9);

	// Release kernel events.
	clReleaseEvent(kernel_event[0]);

	// Verify results.
	verify();

	file.close();
	if (!file.is_open())
		printf("Successfully closed file\n");
}

void verify() {
  printf("Verifying\n");

	for (unsigned i = 0; i < NUM_IMAGES; i++) {
		int max_idx = 0;
		// Read the results.
		cl_int status = clEnqueueReadBuffer(queue[0], output_buf[i], CL_TRUE,
				0, Y_height * Y_width * sizeof(float), output, 0, NULL, NULL);
		checkError(status, "Failed to read output matrix");

		for (unsigned j = 1; j < Y_y; j++){
				float tmp;
				if (output[max_idx] < output[j])
					max_idx = j;
		}

		if (max_idx == TEST_NUM){
			correct_count++;
			//printf("The linear classifier reads CORRECTLY a %d\n", TEST_NUM);
		} else {
			printf("The linear classifier reads it WRONG, it read a %d\n", max_idx);
		}
	}
}

// Free the resources allocated during initialization
void cleanup() {
  for(unsigned i = 0; i < num_devices; i++) {
    if(kernel && kernel[i]) {
      clReleaseKernel(kernel[i]);
    }
    if(queue && queue[i]) {
      clReleaseCommandQueue(queue[i]);
    }

		for (unsigned j = 0; j < NUM_IMAGES; j++) {
			if(input_x_buf[j]) {
				clReleaseMemObject(input_x_buf[j]);
			}
			if(output_buf[j]) {
	      clReleaseMemObject(output_buf[j]);
	    }
		}
  }

  if(program) {
    clReleaseProgram(program);
  }
  if(context) {
    clReleaseContext(context);
  }

	for (int i = 0; i < NUM_IMAGES; i++) {
		alignedFree(input_x[i]);
	}
}
