#include <sys/stat.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <math.h>
#include "CL/opencl.h"
#include "AOCLUtils/aocl_utils.h"
#include "one_layer_nn.h"

using namespace aocl_utils;

#define NUM_IMAGES 	500
#define TEST_NUM 		1

unsigned int correct_count = 0;

// OpenCL runtime configuration
cl_platform_id platform = NULL;
unsigned num_devices = 0;
scoped_array<cl_device_id> device; // num_devices elements
cl_context context = NULL;
scoped_array<cl_command_queue> queue; // num_devices elements
cl_program program = NULL;
scoped_array<cl_kernel> kernel; // num_devices elements/ num_kernels

scoped_array<cl_mem> input_x_buf;
scoped_array<cl_mem> z_buf;
scoped_array<cl_mem> output_buf;

float* input_x[NUM_IMAGES];
float output[Y_y];
float Z[theta1_x];

// Problem data
unsigned X_height = X_y;
unsigned X_width  = X_x;
unsigned Y_height = Y_y;
unsigned Y_width  = Y_x;

// Open MNIST dataset file
std::ifstream file("../../mnist_dataset/coursera/data1", std::ios::in | std::ios::binary);

// Open time log file
std::ofstream log_file("logs/time_log_std.log");

// Tot time
cl_ulong tot_kernel_time_ns_hidden = 0;
cl_ulong tot_kernel_time_ns_output = 0;
double total_time;

// Kernels.
enum Kernel {
  KERNEL_HIDDEN_LAYER = 0,
  KERNEL_OUTPUT_LAYER,

  NUM_KERNELS
};

// Function prototypes
void read_mnist();
bool init_opencl();
void run();
void verify();
void cleanup();

int main(void) {
  // Read the MNIST data
  read_mnist();

	// Initialize OpenCL.
	if (!init_opencl()) {
		return -1;
	}

	// Run the kernel
	run();

	// Free the resources allocated
	cleanup();

	// Calculate accuracy
	float accuracy = ((float)correct_count/NUM_IMAGES)*100;
	printf("Accuracy: %f\n", accuracy);
	return 0;
}

/////////// HELPER FUNCTIONS //////////

// Reads NUM_IMAGES from coursera MNIST file
void read_mnist() {
  for (int i = 0; i < NUM_IMAGES; i++) {
    float buffer[IMG_SIZE+1];

    if (file.is_open()){
      file.read((char*)(buffer+1), sizeof(float)*IMG_SIZE); // values are floats here
    } else {
  		printf("Could not read file\n");
  		file.close();
  	}
    buffer[0] = 1;

    // 64 byte aligned malloc for DMA
    input_x[i] = (float*)alignedMalloc(sizeof(float)*IMG_SIZE+1);
    memcpy(input_x[i], buffer, sizeof(float)*IMG_SIZE+1);
  }
}

// Initializes the OpenCL objects
bool init_opencl() {
  cl_int status;

  printf("Initializing OpenCL\n");

  if (!setCwdToExeDir()) {
    return false;
  }

  // Get the OpenCL platform
  platform = findPlatform("Altera");
  if (platform == NULL) {
    printf("ERROR: Unable to find Altera OpenCL platform.\n");
  }

  // Query the available OpenCL device.
  device.reset(getDevices(platform, CL_DEVICE_TYPE_ALL, &num_devices));
  printf("Platform: %s\n", getPlatformName(platform).c_str());
  printf("Using %d device(s)\n", num_devices);
  for (unsigned i = 0; i < num_devices; ++i) {
    printf("  %s\n", getDeviceName(device[i]).c_str());
  }

  // Create the context
  context = clCreateContext(NULL, num_devices, device, &oclContextCallback, NULL, &status);
  checkError(status, "Failed to create context");

  // Create the program for all device. Use the first device as the
  // representative device (assuming all device are of the same type)
  std::string binary_file = getBoardBinaryFile("one_layer_nn", device[0]);
  printf("Using AOCX: %s\n", binary_file.c_str());
  program = createProgramFromBinary(context, binary_file.c_str(), device , num_devices);

  // Build the program that was just created
  status = clBuildProgram(program, 0, NULL, "", NULL, NULL);
  checkError(status, "Failed to build program");

  // Create per-device objects
  queue.reset(num_devices);
  kernel.reset(NUM_KERNELS);
  input_x_buf.reset(num_devices*NUM_IMAGES);
  z_buf.reset(num_devices);
  output_buf.reset(num_devices*NUM_IMAGES);

  // Command queue
  queue[0] = clCreateCommandQueue(context, device[0], CL_QUEUE_PROFILING_ENABLE, &status);
  checkError(status, "Failed to create command queue");

  // Kernels
  const char *kernel_name_hidden = "one_layer_nn_hidden_layer";
  kernel[KERNEL_HIDDEN_LAYER] = clCreateKernel(program, kernel_name_hidden, &status);
  checkError(status, "Failed to create hidden layer kernel");

  const char *kernel_name_output = "one_layer_nn_output_layer";
  kernel[KERNEL_OUTPUT_LAYER] = clCreateKernel(program, kernel_name_output, &status);
  checkError(status, "Failed to create output layer kernel");

  for (unsigned i = 0; i < NUM_IMAGES; i++) {
    // Create input and output buffers
    input_x_buf[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_BANK_2_ALTERA,
        X_height * X_width * sizeof(float), NULL, &status);
    checkError(status, "Failed to create buffer for input");

    output_buf[i] = clCreateBuffer(context, CL_MEM_WRITE_ONLY | CL_MEM_BANK_1_ALTERA,
        Y_height * Y_width * sizeof(float), NULL, &status);
    checkError(status, "Failed to create buffer for output");

    // Fill input with all input images
    status = clEnqueueWriteBuffer(queue[0], input_x_buf[i], CL_TRUE, // CL_FALSE
      0, X_width * X_height * sizeof(float), input_x[i], 0, NULL, NULL);
    checkError(status, "Failed to transfer input X");
  }

  // Create temporary buffer z_buf
  z_buf[0] = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_BANK_2_ALTERA,
      theta1_y * X_width * sizeof(float), NULL, &status);
  checkError(status, "Failed to create buffer for z");

  return true;
}

void run() {
  cl_int status;

  // Launch kernels.
  // This is the portion of time that we'll be measuring for throughput
  // benchmarking.
  scoped_array<cl_event> hidden_layer_event(num_devices);
  scoped_array<cl_event> output_layer_event(num_devices);

  const double start_time = getCurrentTimestamp();

	// Set kernel arguments.

  // Set hidden layer output argument
  status = clSetKernelArg(kernel[KERNEL_HIDDEN_LAYER], 0, sizeof(cl_mem), &z_buf[0]); // output
  checkError(status, "Failed to set argument %d", 0);

  // Set output layer input argument
  status = clSetKernelArg(kernel[KERNEL_OUTPUT_LAYER], 1, sizeof(cl_mem), &z_buf[0]); // input
  checkError(status, "Failed to set argument %d", 1);

	// Enqueue kernel.
	// Use a global work size corresponding to the size of the output matrix.
	// Each work-item computes the result for one value of the output matrix,
	// so the global work size has the same dimensions as the output matrix.
	//
	// Events are used to ensure that the kernel is not launched until
	// the writes to the input buffers have completed.
	const size_t global_work_size_hidden[1] = {theta1_y}; // 25
	const size_t local_work_size_hidden[1]  = {theta1_y};

  const size_t global_work_size_output[1] = {theta2_y}; // 10
	const size_t local_work_size_output[1]  = {theta2_y};

	printf("Launching for device %d (global size: %zd)\n", 0, global_work_size_hidden[0]);

	for(unsigned i = 0; i < NUM_IMAGES; i++) {
    // Set hidden layer input argument
    status = clSetKernelArg(kernel[KERNEL_HIDDEN_LAYER], 1, sizeof(cl_mem), &input_x_buf[i]); // input
    checkError(status, "Failed to set argument %d", 1);
    // Set output layer output argument
    status = clSetKernelArg(kernel[KERNEL_OUTPUT_LAYER], 0, sizeof(cl_mem), &output_buf[i]); // output
    checkError(status, "Failed to set argument %d", 0);

    // Execute hidden layer
		status = clEnqueueNDRangeKernel(queue[0], kernel[KERNEL_HIDDEN_LAYER], 1, NULL,
				global_work_size_hidden, local_work_size_hidden, 0, NULL, &hidden_layer_event[0]);
		checkError(status, "Failed to launch kernel");

    // Execute output layer
    status = clEnqueueNDRangeKernel(queue[0], kernel[KERNEL_OUTPUT_LAYER], 1, NULL,
        global_work_size_output, local_work_size_output, 0, NULL, &output_layer_event[0]);
    checkError(status, "Failed to launch kernel");

	}

  // Wait for all kernels to finish / flush the queue
  clWaitForEvents(num_devices, output_layer_event);

  // Get kernel times using the OpenCL event profiling API.
  cl_ulong time_ns_hidden = getStartEndTime(hidden_layer_event[0]);
  //printf("Hidden layer kernel time (device %d): %0.3f ms\n", 0, double(time_ns_hidden) * 1e-6);
  tot_kernel_time_ns_hidden += time_ns_hidden;

  cl_ulong time_ns_output = getStartEndTime(output_layer_event[0]);
  //printf("Output layer kernel time (device %d): %0.3f ms\n", 0, double(time_ns_output) * 1e-6);
  tot_kernel_time_ns_output += time_ns_output;

  const double end_time = getCurrentTimestamp();
	total_time = end_time - start_time;
  cl_ulong tot_kernel_time_ns = tot_kernel_time_ns_hidden + tot_kernel_time_ns_output;

	// Wall-clock time taken.
	printf("\nTime: %0.3f ms\n", total_time * 1e3);

  // Write to timing log file
	log_file << "Total time: " << std::setprecision(3) <<  total_time * 1e3 << "ms\n";
	log_file << "\nTotal kernel time: " << std::setprecision(3) <<  tot_kernel_time_ns*NUM_IMAGES * 1e-6 << "ms\n";
  log_file << "\nKernel time: " << std::setprecision(3) <<  tot_kernel_time_ns * 1e-6 << "ms\n";
	log_file << "\nHidden layer kernel time: " << std::setprecision(3) <<  tot_kernel_time_ns_hidden * 1e-6 << "ms\n";
  log_file << "\nOutput layer kernel time: " << std::setprecision(3) <<  tot_kernel_time_ns_output * 1e-6 << "ms\n";
	log_file.close();

  // Compute the throughput (GFLOPS).
  // There are Y_width * Y_height output values, with each value
  // computed using W_width multiplies and adds.
  const float flops = (float)(2.0f * Y_width * Y_height * X_height * NUM_IMAGES / total_time); // not correct
  printf("\nThroughput: %0.2f GFLOPS\n\n", flops * 1e-9);

  // Release kernel events.
  clReleaseEvent(hidden_layer_event[0]);
  clReleaseEvent(output_layer_event[0]);

  // Verify results.
  verify();

	file.close();
	if (!file.is_open())
		printf("Successfully closed file\n");
}

void verify() {
  printf("Verifying\n");

  for (unsigned i = 0; i < NUM_IMAGES; i++) {
    int max_idx = 0;
    // Read the result.
    cl_int status = clEnqueueReadBuffer(queue[0], output_buf[i], CL_TRUE,
        0, Y_height * Y_width * sizeof(float), output, 0, NULL, NULL);
    checkError(status, "Failed to read output matrix");

    for (int i=1; i < Y_y; i++){
        float tmp;
        if (output[max_idx] < output[i])
          max_idx = i;
    }

    if (max_idx == TEST_NUM-1){ // idx 9 is 0 here
      correct_count++;
      //printf("The linear classifier reads CORRECTLY a %d\n", TEST_NUM);
    } else {
      printf("The linear classifier reads it WRONG, it read a %d\n", max_idx+1);
    }
  }
}

// Free the resources allocated during initialization
void cleanup() {
  for(unsigned i = 0; i < NUM_KERNELS; i++) {
    if(kernel && kernel[i]) {
      clReleaseKernel(kernel[i]);
    }
  }
  if(queue && queue[0]) {
    clReleaseCommandQueue(queue[0]);
  }
  if(z_buf[0]) {
    clReleaseMemObject(z_buf[0]);
  }
  for (unsigned i = 0; i < NUM_IMAGES; i++) {
    if(input_x_buf[i]) {
      clReleaseMemObject(input_x_buf[i]);
    }
    if(output_buf) {
      clReleaseMemObject(output_buf[i]);
    }
  }

  if(program) {
    clReleaseProgram(program);
  }
  if(context) {
    clReleaseContext(context);
  }

  for (int i = 0; i < NUM_IMAGES; i++) {
		alignedFree(input_x[i]);
	}
}
