#include <sys/stat.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <math.h>
#include "CL/opencl.h"
#include "AOCLUtils/aocl_utils.h"
#include "two_layers_nn.h"
#include "../inc/theta1_weights_2h_1d.h"
#include "../inc/theta2_weights_2h_1d.h"
#include "../inc/theta3_weights_2h_1d.h"

using namespace aocl_utils;

#define NUM_IMAGES 	500
#define TEST_NUM 		3

// Kernels.
enum Kernel {
  KERNEL_H1_LAYER = 0,
  KERNEL_H2_LAYER,
  KERNEL_OUTPUT_LAYER,

  NUM_KERNELS
};

unsigned int correct_count = 0;

// OpenCL runtime configuration
cl_platform_id platform = NULL;
unsigned num_devices = 0;
scoped_array<cl_device_id> device; // num_devices elements
cl_context context = NULL;
scoped_array<cl_command_queue> queue; // num_devices elements
cl_program program = NULL;
scoped_array<cl_kernel> kernel; // num_devices elements/ num_kernels

scoped_array<cl_mem> weights_buf;
scoped_array<cl_mem> input_x_buf;
scoped_array<cl_mem> z1_buf;
scoped_array<cl_mem> z2_buf;
scoped_array<cl_mem> output_buf;

float* input_x[NUM_IMAGES];
float* output;
float* weights[NUM_KERNELS];

// Problem data
unsigned X_height = X_y;
unsigned X_width  = X_x;
unsigned Y_height = Y_y;
unsigned Y_width  = Y_x;

unsigned theta1_width = theta1_x;
unsigned theta2_width = theta2_x;
unsigned theta3_width = theta3_x;

// Open MNIST dataset file
std::ifstream file("../../mnist_dataset/coursera/data3", std::ios::in | std::ios::binary);

// Open time log file
std::ofstream log_file("logs/time_log_std_tmp.log");

// Tot time
cl_ulong tot_kernel_time_ns_h1 = 0;
cl_ulong tot_kernel_time_ns_h2 = 0;
cl_ulong tot_kernel_time_ns_output = 0;
double total_time;

// Function prototypes
void read_mnist();
bool init_opencl();
void run();
void verify();
void cleanup();

int main(void) {
  // Read the MNIST data
  read_mnist();

	// Initialize OpenCL.
	if (!init_opencl()) {
		return -1;
	}

	// Run the kernel
	run();

	// Free the resources allocated
	cleanup();

	// Calculate accuracy
	float accuracy = ((float)correct_count/NUM_IMAGES)*100;
	printf("Accuracy: %f\n", accuracy);
	return 0;
}

/////////// HELPER FUNCTIONS //////////

void read_mnist() {
  for (int i = 0; i < NUM_IMAGES; i++) {
    float buffer[IMG_SIZE];

    if (file.is_open()){
      file.read((char*)(buffer), sizeof(float)*IMG_SIZE); // values are floats here
    } else {
  		printf("Could not read file\n");
  		file.close();
  	}

    // 64 byte aligned malloc for DMA
    input_x[i] = (float*)alignedMalloc(sizeof(float)*IMG_SIZE);
    memcpy(input_x[i], buffer, sizeof(float)*IMG_SIZE);
  }

  // 64 byte aligned malloc for DMA
  output = (float*)alignedMalloc(sizeof(float)*Y_y);

  weights[KERNEL_H1_LAYER] = (float*)alignedMalloc(sizeof(float)*theta1_y*theta1_x);
  weights[KERNEL_H2_LAYER] = (float*)alignedMalloc(sizeof(float)*theta2_y*theta2_x);
  weights[KERNEL_OUTPUT_LAYER] = (float*)alignedMalloc(sizeof(float)*theta3_y*theta3_x);

  memcpy(weights[KERNEL_H1_LAYER], Theta1, sizeof(float)*theta1_y*theta1_x);
  memcpy(weights[KERNEL_H2_LAYER], Theta2, sizeof(float)*theta2_y*theta2_x);
  memcpy(weights[KERNEL_OUTPUT_LAYER], Theta3, sizeof(float)*theta3_y*theta3_x);
}

// Initializes the OpenCL objects
bool init_opencl() {
  cl_int status;

  printf("Initializing OpenCL\n");

  if (!setCwdToExeDir()) {
    return false;
  }

  // Get the OpenCL platform
  platform = findPlatform("Altera");
  if (platform == NULL) {
    printf("ERROR: Unable to find Altera OpenCL platform.\n");
  }

  // Query the available OpenCL device.
  device.reset(getDevices(platform, CL_DEVICE_TYPE_ALL, &num_devices));
  printf("Platform: %s\n", getPlatformName(platform).c_str());
  printf("Using %d device(s)\n", num_devices);
  for (unsigned i = 0; i < num_devices; ++i) {
    printf("  %s\n", getDeviceName(device[i]).c_str());
  }

  // Create the context
  context = clCreateContext(NULL, num_devices, device, &oclContextCallback, NULL, &status);
  checkError(status, "Failed to create context");

  // Create the program for all device. Use the first device as the
  // representative device (assuming all device are of the same type)
  std::string binary_file = getBoardBinaryFile("two_layers_nn_simd_max", device[0]);
  printf("Using AOCX: %s\n", binary_file.c_str());
  program = createProgramFromBinary(context, binary_file.c_str(), device , num_devices);

  // Build the program that was just created
  status = clBuildProgram(program, 0, NULL, "", NULL, NULL);
  checkError(status, "Failed to build program");

  // Create per-device objects
  queue.reset(num_devices);
  kernel.reset(NUM_KERNELS);
  weights_buf.reset(num_devices*NUM_KERNELS);
  input_x_buf.reset(num_devices*NUM_IMAGES);
  z1_buf.reset(num_devices);
  z2_buf.reset(num_devices);
  output_buf.reset(num_devices*NUM_IMAGES);

  const unsigned num_block_rows = Y_height; // Y ?? / BLOCK_SIZE;

  // Command queue
  queue[0] = clCreateCommandQueue(context, device[0], CL_QUEUE_PROFILING_ENABLE, &status);
  checkError(status, "Failed to create command queue");


  const char *kernel_name_100 = "hidden_layer_nn_100";
  kernel[KERNEL_H1_LAYER] = clCreateKernel(program, kernel_name_100, &status);
  checkError(status, "Failed to create kernel, layer: %d", KERNEL_H1_LAYER);

  const char *kernel_name_25 = "hidden_layer_nn_25";
  kernel[KERNEL_H2_LAYER] = clCreateKernel(program, kernel_name_25, &status);
  checkError(status, "Failed to create kernel, layer: %d", KERNEL_H2_LAYER);

  const char *kernel_name_10 = "hidden_layer_nn_10";
  kernel[KERNEL_OUTPUT_LAYER] = clCreateKernel(program, kernel_name_10, &status);
  checkError(status, "Failed to create kernel, layer: %d", KERNEL_OUTPUT_LAYER);

  // Create weight buffers
	weights_buf[KERNEL_H1_LAYER] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_BANK_1_ALTERA,
      theta1_y * theta1_x * sizeof(float), NULL, &status);
  checkError(status, "Failed to create buffer for weights Theta1");

  weights_buf[KERNEL_H2_LAYER] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_BANK_1_ALTERA,
      theta2_y * theta2_x * sizeof(float), NULL, &status);
  checkError(status, "Failed to create buffer for weights Theta2");

  weights_buf[KERNEL_OUTPUT_LAYER] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_BANK_1_ALTERA,
      theta3_y * theta3_x * sizeof(float), NULL, &status);
  checkError(status, "Failed to create buffer for weights Theta3");

  // Write weight values to the weight buffers.
  // Can be done here since the weight values does not change.
  status = clEnqueueWriteBuffer(queue[0], weights_buf[KERNEL_H1_LAYER], CL_TRUE, // CL_FALSE
    0, theta1_y * theta1_x * sizeof(float), weights[KERNEL_H1_LAYER], 0, NULL, NULL);
  checkError(status, "Failed to transfer input X");

  status = clEnqueueWriteBuffer(queue[0], weights_buf[KERNEL_H2_LAYER], CL_TRUE, // CL_FALSE
    0, theta2_y * theta2_x * sizeof(float), weights[KERNEL_H2_LAYER], 0, NULL, NULL);
  checkError(status, "Failed to transfer input X");

  status = clEnqueueWriteBuffer(queue[0], weights_buf[KERNEL_OUTPUT_LAYER], CL_TRUE, // CL_FALSE
    0, theta3_y * theta3_x * sizeof(float), weights[KERNEL_OUTPUT_LAYER], 0, NULL, NULL);
  checkError(status, "Failed to transfer input X");

  for (unsigned i = 0; i < NUM_IMAGES; i++) {
    // Create input and output buffers
    input_x_buf[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_BANK_2_ALTERA,
        X_height * X_width * sizeof(float), NULL, &status);
    checkError(status, "Failed to create buffer for input");

    output_buf[i] = clCreateBuffer(context, CL_MEM_WRITE_ONLY | CL_MEM_BANK_1_ALTERA,
        Y_height * Y_width * sizeof(float), NULL, &status);
    checkError(status, "Failed to create buffer for output");

    // Fill input with all input images
    status = clEnqueueWriteBuffer(queue[0], input_x_buf[i], CL_TRUE, // CL_FALSE
			0, X_width * X_height * sizeof(float), input_x[i], 0, NULL, NULL);
		checkError(status, "Failed to transfer input X");
  }

  // Create temporary buffers
  z1_buf[0] = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_BANK_2_ALTERA,
      theta1_y * sizeof(float), NULL, &status);
  checkError(status, "Failed to create buffer for z");

  z2_buf[0] = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_BANK_2_ALTERA,
      theta2_y * sizeof(float), NULL, &status);
  checkError(status, "Failed to create buffer for z");

  return true;
}

void run() {
  cl_int status;

  // Launch kernels.
  // This is the portion of time that we'll be measuring for throughput
  // benchmarking.
  scoped_array<cl_event> kernel_h1_event(num_devices);
  scoped_array<cl_event> kernel_h2_event(num_devices);
  scoped_array<cl_event> kernel_output_event(num_devices);

  const double start_time = getCurrentTimestamp();

	// Set kernel arguments.
  // Hidden layer 1 kernel
  unsigned argi = 0;

  status = clSetKernelArg(kernel[KERNEL_H1_LAYER], argi++, sizeof(cl_mem), &z1_buf[0]); // output
  checkError(status, "Failed to set argument %d", argi - 1);

  status = clSetKernelArg(kernel[KERNEL_H1_LAYER], argi++, sizeof(cl_mem), &weights_buf[0]); //
  checkError(status, "Failed to set argument %d", argi - 1);

  argi++; // input argument set in the execution loop below

  status = clSetKernelArg(kernel[KERNEL_H1_LAYER], argi++, sizeof(theta1_width), &theta1_width);
  checkError(status, "Failed to set argument %d", argi - 1);

  // Hidden layer 2 kernel
  argi = 0;

  status = clSetKernelArg(kernel[KERNEL_H2_LAYER], argi++, sizeof(cl_mem), &z2_buf[0]); // output
  checkError(status, "Failed to set argument %d", argi - 1);

  status = clSetKernelArg(kernel[KERNEL_H2_LAYER], argi++, sizeof(cl_mem), &weights_buf[1]); //
  checkError(status, "Failed to set argument %d", argi - 1);

  status = clSetKernelArg(kernel[KERNEL_H2_LAYER], argi++, sizeof(cl_mem), &z1_buf[0]); // input
  checkError(status, "Failed to set argument %d", argi - 1);

  status = clSetKernelArg(kernel[KERNEL_H2_LAYER], argi++, sizeof(theta2_width), &theta2_width);
  checkError(status, "Failed to set argument %d", argi - 1);

  // Output layer kernel
  argi = 0;

  argi++; // output argument set in the execution loop below;

  status = clSetKernelArg(kernel[KERNEL_OUTPUT_LAYER], argi++, sizeof(cl_mem), &weights_buf[2]); //
  checkError(status, "Failed to set argument %d", argi - 1);

  status = clSetKernelArg(kernel[KERNEL_OUTPUT_LAYER], argi++, sizeof(cl_mem), &z2_buf[0]); // input
  checkError(status, "Failed to set argument %d", argi - 1);

  status = clSetKernelArg(kernel[KERNEL_OUTPUT_LAYER], argi++, sizeof(theta3_width), &theta3_width);
  checkError(status, "Failed to set argument %d", argi - 1);


	// Enqueue kernel.
	// Use a global work size corresponding to the size of the output matrix.
	// Each work-item computes the result for one value of the output matrix,
	// so the global work size has the same dimensions as the output matrix.
	//
	// Events are used to ensure that the kernel is not launched until
	// the writes to the input buffers have completed.
	const size_t global_work_size_h1[1] = {theta1_y}; // 100
	const size_t local_work_size_h1[1]  = {theta1_y};

  const size_t global_work_size_h2[1] = {theta2_y}; // 25
  const size_t local_work_size_h2[1]  = {theta2_y};

  const size_t global_work_size_output[1] = {theta3_y}; // 10
	const size_t local_work_size_output[1]  = {theta3_y};

	printf("Launching for device %d (global sizes:\nh1: %zd)\nh2: %zd)\noutput: %zd)\n",
          0, global_work_size_h1[0], global_work_size_h2[0], global_work_size_output[0]);

	for(unsigned i = 0; i < NUM_IMAGES; i++) {
    // Set remaining arguments
    status = clSetKernelArg(kernel[KERNEL_H1_LAYER], 2, sizeof(cl_mem), &input_x_buf[i]); // input
    checkError(status, "Failed to set argument %d", 2);

    status = clSetKernelArg(kernel[KERNEL_OUTPUT_LAYER], 0, sizeof(cl_mem), &output_buf[i]); // output
    checkError(status, "Failed to set argument %d", 0);

    /****************** Hidden layer 1 *********************/
		status = clEnqueueNDRangeKernel(queue[0], kernel[KERNEL_H1_LAYER], 1, NULL,
				global_work_size_h1, local_work_size_h1, 0, NULL, &kernel_h1_event[0]);
		checkError(status, "Failed to launch kernel");

    /****************** Hidden layer 2 *********************/
    status = clEnqueueNDRangeKernel(queue[0], kernel[KERNEL_H2_LAYER], 1, NULL,
        global_work_size_h2, local_work_size_h2, 0, NULL, &kernel_h2_event[0]);
    checkError(status, "Failed to launch kernel");


    /******************* Output layer **********************/
    status = clEnqueueNDRangeKernel(queue[0], kernel[KERNEL_OUTPUT_LAYER], 1, NULL,
        global_work_size_output, local_work_size_output, 0, NULL, &kernel_output_event[0]);
    checkError(status, "Failed to launch kernel");

	}

  // Wait for all kernels to finish.
  clWaitForEvents(num_devices, kernel_output_event);

  cl_ulong time_ns_h1 = getStartEndTime(kernel_h1_event[0]);
  //printf("Hidden layer kernel time (device %d): %0.3f ms\n", 0, double(time_ns_h1) * 1e-6);
  tot_kernel_time_ns_h1 += time_ns_h1;
  cl_ulong time_ns_h2 = getStartEndTime(kernel_h2_event[0]);
  //printf("Hidden layer kernel time (device %d): %0.3f ms\n", 0, double(time_ns_h2) * 1e-6);
  tot_kernel_time_ns_h2 += time_ns_h2;
  cl_ulong time_ns_output = getStartEndTime(kernel_output_event[0]);
  //printf("Output layer kernel time (device %d): %0.3f ms\n", 0, double(time_ns_output) * 1e-6);
  tot_kernel_time_ns_output += time_ns_output;

  const double end_time = getCurrentTimestamp();
  const double total_time = end_time - start_time;
  cl_ulong tot_kernel_time_ns = tot_kernel_time_ns_h1 + tot_kernel_time_ns_h2 + tot_kernel_time_ns_output;

  // Wall-clock time taken.
  printf("\nTime: %0.3f ms\n", total_time * 1e3);

  // Write to timing log file
  log_file << "Total time: " << std::setprecision(3) <<  total_time * 1e3 << "ms\n";
  log_file << "\nTotal kernel time: " << std::setprecision(3) <<  tot_kernel_time_ns*NUM_IMAGES * 1e-6 << "ms\n";
  log_file << "\nKernel time: " << std::setprecision(3) <<  tot_kernel_time_ns * 1e-6 << "ms\n";
  log_file << "\nHidden layer 1 kernel time: " << std::setprecision(3) <<  tot_kernel_time_ns_h1 * 1e-6 << "ms\n";
  log_file << "\nHidden layer 2 kernel time: " << std::setprecision(3) <<  tot_kernel_time_ns_h2 * 1e-6 << "ms\n";
  log_file << "\nOutput layer kernel time: " << std::setprecision(3) <<  tot_kernel_time_ns_output * 1e-6 << "ms\n";
  log_file.close();

  // Compute the throughput (GFLOPS).
  // There are Y_width * Y_height output values, with each value
  // computed using W_width multiplies and adds.
  const float flops = (float)(2.0f * Y_width * Y_height * X_height * NUM_IMAGES/ total_time); // not correct
  printf("\nThroughput: %0.2f GFLOPS\n\n", flops * 1e-9);

  // Release kernel events.
  clReleaseEvent(kernel_h1_event[0]);
  clReleaseEvent(kernel_h2_event[0]);
  clReleaseEvent(kernel_output_event[0]);

  // Verify results.
  verify();

	file.close();
	if (!file.is_open())
		printf("Successfully closed file\n");
}

void verify() {
  printf("Verifying\n");

  for (unsigned i = 0; i < NUM_IMAGES; i++) {
    int max_idx = 0;
    // Read the result.
    cl_int status = clEnqueueReadBuffer(queue[0], output_buf[i], CL_TRUE,
        0, Y_height * Y_width * sizeof(float), output, 0, NULL, NULL);
    checkError(status, "Failed to read output matrix");

    for (int j=1; j < Y_y; j++){
        float tmp;
        if (output[max_idx] < output[j])
          max_idx = j;
    }

    if (max_idx == TEST_NUM-1){
      correct_count++;
      //printf("The linear classifier reads CORRECTLY a %d\n", TEST_NUM);
    } else if (max_idx == 9) { // idx 9 is 0 here
      correct_count++;
      //printf("The linear classifier reads CORRECTLY a %d\n", TEST_NUM);
    } else {
      printf("The linear classifier reads it WRONG, it read a %d\n", max_idx+1);
    }
  }
}

// Free the resources allocated during initialization
void cleanup() {
  for(unsigned i = 0; i < NUM_KERNELS; ++i) {
    if(kernel && kernel[i]) {
      clReleaseKernel(kernel[i]);
    }
  }
  if(queue && queue[0]) {
    clReleaseCommandQueue(queue[0]);
  }
  if(weights_buf[KERNEL_H1_LAYER]) {
    clReleaseMemObject(weights_buf[KERNEL_H1_LAYER]);
  }
  if(weights_buf[KERNEL_H2_LAYER]) {
    clReleaseMemObject(weights_buf[KERNEL_H2_LAYER]);
  }
  if(weights_buf[KERNEL_OUTPUT_LAYER]) {
    clReleaseMemObject(weights_buf[KERNEL_OUTPUT_LAYER]);
  }
  if(z1_buf[0]) {
    clReleaseMemObject(z1_buf[0]);
  }
  if(z2_buf[0]) {
    clReleaseMemObject(z2_buf[0]);
  }
  for (unsigned i = 0; i < NUM_IMAGES; i++) {
    if(input_x_buf[i]) {
      clReleaseMemObject(input_x_buf[i]);
    }
    if(output_buf) {
      clReleaseMemObject(output_buf[i]);
    }
  }

  if(program) {
    clReleaseProgram(program);
  }
  if(context) {
    clReleaseContext(context);
  }

  for (int i = 0; i < NUM_IMAGES; i++) {
    alignedFree(input_x[i]);
  }
}
